
#include <XSUtil/FunctionUtils/xsFortran.h>
#include <XSUtil/FunctionUtils/FunctionUtility.h>
#include <XSUtil/Utils/XSutility.h>
#include <XSUtil/Utils/IosHolder.h>
#include <XSUtil/Utils/XSstream.h>
#include <XSUtil/Numerics/CosmologyFunction.h>
#include <XSUtil/Numerics/Gamma.h>
#include <cfortran.h>
#include <cstdlib>
#include <string>

FCALLSCFUN0(STRING,FGDATD,FGDATD,fgdatd)
FCALLSCFUN0(STRING,FGMODF,FGMODF,fgmodf)
FCALLSCFUN0(STRING,FGXSCT,FGXSCT,fgxsct)
FCALLSCFUN0(STRING,FGSOLR,FGSOLR,fgsolr)
FCALLSCFUN1(FLOAT,FGABND,FGABND,fgabnd,STRING)
FCALLSCFUN1(STRING,FGMSTR,FGMSTR,fgmstr,STRING)
FCALLSCSUB1(FPDATD,FPDATD,fpdatd,STRING)
FCALLSCSUB2(FPSOLR,FPSOLR,fpsolr,STRING,PINT)
FCALLSCSUB2(FPXSCT,FPXSCT,fpxsct,STRING,PINT)
FCALLSCSUB2(FPMSTR,FPMSTR,fpmstr,STRING,STRING)
FCALLSCSUB3(FPSLFL,FPSLFL,fpslfl,FLOATV,INT,PINT)
FCALLSCSUB0(FNINIT,FNINIT,fninit)
FCALLSCFUN2(INT,xs_write,XWRITE,xwrite,STRING,INT)
FCALLSCFUN3(INT,xs_read,XCREAD,xcread,STRING,PSTRING,PINT)
FCALLSCFUN0(FLOAT,csmgq0,CSMGQ0,csmgq0)
FCALLSCFUN0(FLOAT,csmgh0,CSHMH0,csmgh0)
FCALLSCFUN0(FLOAT,csmgl0,CSMGL0,csmgl0)
FCALLSCSUB1(csmpq0,CSMPQ0,csmpq0,FLOAT)
FCALLSCSUB1(csmph0,CSMPH0,csmph0,FLOAT)
FCALLSCSUB1(csmpl0,CSMPL0,csmpl0,FLOAT)
FCALLSCFUN3(FLOAT,fzsq,FZSQ,fzsq,FLOAT,FLOAT,FLOAT)
FCALLSCFUN0(INT,FGCHAT,FGCHAT,fgchat)
FCALLSCSUB1(FPCHAT,FPCHAT,fpchat,INT)
FCALLSCSUB2(xs_getChat,XGTCHT,xgtcht,PINT,PINT)
FCALLSCFUN2(INT,xs_getVersion,XGVERS,xgvers,PSTRING,INT)
FCALLSCFUN1(FLOAT,xs_erf,ERF,erf,FLOAT)
FCALLSCFUN1(FLOAT,xs_erfc,ERFC,erfc,FLOAT)
FCALLSCFUN2(FLOAT,gammap,GAMMAP,gammap,FLOAT,FLOAT)
FCALLSCFUN2(FLOAT,gammq,GAMMQ,gammq,FLOAT,FLOAT)

char* FGDATD()
{
        return const_cast<char*>(FunctionUtility::managerPath().c_str());
        
} 

char* FGMODF()
{
        return const_cast<char*>(FunctionUtility::modelDataPath().c_str());       
} 

char* FGXSCT()
{
        return const_cast<char*>(FunctionUtility::XSECT().c_str());
        
}

char* FGSOLR()
{
        return const_cast<char*>(FunctionUtility::ABUND().c_str());
        
}

float FGABND(char* element)
{
        string elt(element);
        return  FunctionUtility::getAbundance(elt);      
}


char* FGMSTR(char* dname)
{
        static string value;
        value = FunctionUtility::getModelString(string(dname));
        if (value == FunctionUtility::NOT_A_KEY())
        {
           value.erase();
        }
        return const_cast<char*>(value.c_str());      
}

void FPDATD(const char* dataDir)
{
   FunctionUtility::managerPath(string(dataDir));
}

void FPSOLR(const char* table, int* ierr)
{
   string tableName = string(table);
   tableName = XSutility::lowerCase(tableName);
   if (tableName == "file")
   {
      FunctionUtility::ABUND(tableName);
      *ierr = 0;
   }
   else
   {
      if (FunctionUtility::checkAbund(tableName))
      {
         FunctionUtility::ABUND(tableName);
         *ierr = 0;
      }
      else
      {
         *ierr = 1;
      }
   }
}

void FPXSCT(const char* csection, int* ierr)
{
   string tableName = string(csection);
   tableName = XSutility::lowerCase(tableName);
   if (FunctionUtility::checkXsect(tableName))
   {
      FunctionUtility::XSECT(tableName);
      *ierr = 0;
   }
   else
   {
      *ierr = 1;
   }
}

void FPMSTR(const char* value1, const char* value2)
{
   string svalue1 = XSutility::upperCase(string(value1));
   if (svalue1 == "INITIALIZE")
   {
      FunctionUtility::eraseModelStringDataBase();
   }
   else
   {
      FunctionUtility::setModelString(svalue1, string(value2));
   }
}

void FPSLFL(float rvalue[], int nvalue, int *ierr)
{
   // Load the values of the file solar abundance table.
   *ierr = 0;
   size_t nelem = static_cast<size_t>(nvalue);
   if (nelem != FunctionUtility::NELEMS())
   {
      *ierr = 1;
   } 
   else
   {
      std::vector<float> rvect(nelem);
      for (size_t i=0; i<nelem; ++i)
      {
         rvect[i] = rvalue[i];
      }
      FunctionUtility::abundanceVectors("file",rvect);
   }  
}

void FNINIT()
{
   
   // Initialize the directory for the data files.  Use the
   // XSPEC_MDATA_DIR environment variable if it is set, otherwise
   // look in the manager directory.
   const char* cModelDataLoc = getenv("XSPEC_MDATA_DIR");
   const char* cHeadasLoc = getenv("HEADAS");
   string modelDataLoc;
   string managerLoc;
   
   if (cModelDataLoc)
   {
      modelDataLoc = string(cModelDataLoc);
   }
   else if (cHeadasLoc)
   {
      modelDataLoc = string(cHeadasLoc);
      modelDataLoc += "/../spectral/modelData/";
   }
   else
   {
      *IosHolder::outHolder() << "\n***Error: HEADAS environment variable not set.  Unable to find"
           << "\n  spectral/modelData or spectral/manager directories." << std::endl;
      return; 
   }
   
   if (!cHeadasLoc)
   {
      *IosHolder::outHolder() << "\n***Error: HEADAS environment variable not set.  Unable to find"
           << "\n  spectral/manager directory." << std::endl;
      return; 
   }
   else
   {
      managerLoc = string(cHeadasLoc);
      managerLoc += "/../spectral/manager";
   }
   
   FunctionUtility::modelDataPath(modelDataLoc);
   FunctionUtility::managerPath(managerLoc);
   
   // Read in the abundance and crosssections.dat files.  
   string dummy1, dummy2;
   FunctionUtility::readInitializers(dummy1, dummy2);
   
   int ierr=1;
   
   // Initialize the solar abundance table in use.
   FPSOLR("angr", &ierr);
   if (ierr)
   {
      *IosHolder::outHolder() << "\n***Error: Failed to set Solar abundance table."<<std::endl;
      return;
   }
   
   // Initialize the photoelectric cross-sections in use.
   FPXSCT("bcmc", &ierr);
   if (ierr)
   {
      *IosHolder::outHolder() << "\n***Error: Failed to set photoelectric cross-section"<<std::endl;
      return;
   }
   
   // Initialize the list of model string parameters
   FPMSTR("INITIALIZE"," ");
         
}

// Emulates the xanlib routine FXWRITE.
int xs_write(char *wrtstr,int  idest)
{
  XSstream* xsout = dynamic_cast<XSstream*>(IosHolder::outHolder());
  if (xsout)
  {   
     if (idest < 0 )
     {
           // write only to log file.
           if ( xsout->logChatterLevel() >= std::abs(idest))
           {
                   XSstream::verbose(*xsout,0,std::abs(idest)); 
                   *xsout << wrtstr << std::endl; 
                   XSstream::verbose(*xsout);                
           }

     }
     else
     {
           if ( xsout->maxChatter() >= idest ) *xsout << wrtstr << std::endl;
     } 
  }
  else
  {
     if (FunctionUtility::xwriteChatter() >= std::abs(idest))
        *IosHolder::outHolder() << wrtstr << std::endl;
  }
  return 0;
}

// Emulates XCREAD, etc, but assumes interactive input.
int xs_read(const char *prompt,char * buffer,int* ierr)
{
  XSstream* xsin = dynamic_cast<XSstream*>(IosHolder::inHolder());
  char  bk[] = "/*";
  if (xsin)
  {      
     char* temp_prompt =  new char [strlen(IosHolder::xsPrompt())+1]; 
     XSstream::setPrompter(*xsin,string(prompt));
     *xsin >> buffer;
     XSstream::setPrompter(*xsin,string(temp_prompt));
     delete [] temp_prompt;
  }
  else
  {
     *IosHolder::inHolder() >> buffer;
  }
  
  if( !strncmp(buffer, bk, strlen(bk)) )
  {
        *ierr = -1;
        return 0;
  }

  *ierr = 0;
  
  return 0;
}

float csmgq0()
{
   return FunctionUtility::getq0();
}

float csmgh0()
{
   return FunctionUtility::getH0();
}

float csmgl0()
{
   return FunctionUtility::getlambda0();
}

void csmpq0(float q0)
{
   FunctionUtility::setq0(q0);
}

void csmph0(float H0)
{
   FunctionUtility::setH0(H0);
}

void csmpl0(float lambda0)
{
   FunctionUtility::setlambda0(lambda0);
}

float fzsq(float z, float q0, float lambda)
{
   Numerics::FZSQ val;
   return val(z,q0,lambda);
}

int FGCHAT()
{
   return FunctionUtility::xwriteChatter();
}

void FPCHAT(int chat)
{
   FunctionUtility::xwriteChatter(chat);
}

// Emulates xanlib/xparse's xgtcht, but only when this library is
// linked into xspec's executable.  For external programs this
// simply returns levels = 0.
void xs_getChat(int* cons, int* log)
{
  XSstream* xsout = dynamic_cast<XSstream*>(IosHolder::outHolder());
  if (xsout)
  {
     *cons = xsout->consoleChatterLevel();
     *log = xsout->logChatterLevel();
  }
  else
  {
     *cons = 0;
     *log = 0;
  }   
}

int xs_getVersion(char* buffer, int buffSize)
{
   const string& versStr(XSutility::xs_version());
   const int len = static_cast<int>(versStr.length());
   int isOK = -1;
   if (buffSize > len)
   {
      versStr.copy(buffer, len);
      buffer[len] = 0;
      isOK = 0;
   }
   else if (buffSize > 0)
   {
      versStr.copy(buffer, buffSize-1);
      buffer[buffSize-1] = 0;
   }
   return isOK;
}


float xs_erf(float x)
{
        Numerics::Erf E;
        Real y(x);
        return float(E(y));
}

float xs_erfc(float x)
{
        Numerics::Erfc E;
        Real y(x);
        return float(E(y));
}

float gammap(float a, float x)
{
        Numerics::GammaP G;
        Real y(x);
        Real b(a);
        return float(G(b,y));
}



float gammq(float a, float x)
{
        Numerics::GammaQ G;
        Real y(x);
        Real b(a);
        return float(G(b,y));
}


